if ("undefined" == typeof(wdw_imageEdition)) {
	var { Services } = ChromeUtils.import("resource://gre/modules/Services.jsm");
	var { NetUtil } = ChromeUtils.import("resource://gre/modules/NetUtil.jsm");
	var { FileUtils } = ChromeUtils.import("resource://gre/modules/FileUtils.jsm");
	// var { AppConstants } = ChromeUtils.import("resource://gre/modules/AppConstants.jsm");
	var { cardbookRepository } = ChromeUtils.import("chrome://cardbook/content/cardbookRepository.js");
	
	var loader = Services.scriptloader;
	loader.loadSubScript("chrome://cardbook/content/cardbookWebDAV.js", this);
	
	var wdw_imageEdition = {
		
		setMediaContentForCard: function(aType, aExtension, aValue) {
			document.getElementById(aType + 'URITextBox').value = aValue;
			document.getElementById(aType + 'ExtensionTextBox').value = aExtension;
		},

		clearImageCard: function () {
			document.getElementById('defaultCardImage').src = "";
			document.getElementById('imageForSizing').src = "";
			document.getElementById('imageBox').setAttribute('hidden', 'true');
		},

		displayImageCard: async function (aCard, aDisplayDefault) {
			let dirname = cardbookRepository.cardbookPreferences.getName(aCard.dirPrefId);
			wdw_imageEdition.setMediaContentForCard("logo", aCard.logo.extension, aCard.logo.value);
			wdw_imageEdition.setMediaContentForCard("sound", aCard.sound.extension, aCard.sound.value);
			if (aCard.photo.attachmentId) {
				document.getElementById('photoAttachmentIdTextBox').value = aCard.photo.attachmentId;
			}
			if (aCard.photo.value) {
				let image = {cbid: aCard.dirPrefId+"::"+aCard.uid, dirPrefId: aCard.dirPrefId, extension: aCard.photo.extension, content: aCard.photo.value};
				wdw_imageEdition.resizeImageCard(image, aDisplayDefault);
			} else {
				await cardbookIDBImage.getImage("photo", dirname, aCard.cbid, aCard.fn)
				.then( image => {
					wdw_imageEdition.resizeImageCard(image, aDisplayDefault);
				}).catch( () => {
					if (aDisplayDefault) {
						wdw_imageEdition.resizeImageCard(null, aDisplayDefault);
					} else {
						document.getElementById('imageBox').setAttribute('hidden', 'true');
					}
				});
			}
		},

		resizeImageCard: function (aContent, aDisplayDefault) {
			let myImage = document.getElementById('defaultCardImage');
			let myDummyImage = document.getElementById('imageForSizing');
			let content = 	"";
			if (aContent && aContent.content) {
				content = 'data:image/' + aContent.extension + ';base64,' + aContent.content;
				wdw_imageEdition.setMediaContentForCard("photo", aContent.extension, aContent.content);
			} else {
				content = cardbookRepository.defaultCardImage;
				document.getElementById('photoURITextBox').value = "";
				document.getElementById('photoExtensionTextBox').value = "";
			}

			myImage.src = "";
			myDummyImage.src = "";
			myDummyImage.src = content;
			myDummyImage.onload = function(e) {
				var myImageWidth = 170;
				var myImageHeight = 170;
				if (myDummyImage.width >= myDummyImage.height) {
					widthFound = myImageWidth + "px" ;
					heightFound = Math.round(myDummyImage.height * myImageWidth / myDummyImage.width) + "px" ;
				} else {
					widthFound = Math.round(myDummyImage.width * myImageHeight / myDummyImage.height) + "px" ;
					heightFound = myImageHeight + "px" ;
				}
				myImage.width = widthFound;
				myImage.height = heightFound;
				myImage.src = content;
				document.getElementById('imageBox').removeAttribute('hidden');
			}
			myDummyImage.onerror = function(e) {
				document.getElementById('photoURITextBox').value = "";
				document.getElementById('photoExtensionTextBox').value = "";
				if (aDisplayDefault) {
					wdw_imageEdition.resizeImageCard(null, aDisplayDefault);
				}
			}
		},

		saveImageCard: function () {
			let defaultName = document.getElementById('fnTextBox').value + "." + document.getElementById('photoExtensionTextBox').value.toLowerCase();
			cardbookWindowUtils.callFilePicker("imageSaveTitle", "SAVE", "IMAGES", defaultName, "", wdw_imageEdition.saveImageCardNext);
		},

		saveImageCardNext: async function (aFile) {
			await cardbookRepository.cardbookUtils.writeContentToFile(aFile.path, atob(document.getElementById('photoURITextBox').value), "NOUTF8");
			cardbookRepository.cardbookUtils.formatStringForOutput("imageSavedToFile", [aFile.path]);
		},

		copyImageCard: function () {
			cardbookClipboard.clipboardSetImage(document.getElementById('photoURITextBox').value, document.getElementById('photoExtensionTextBox').value);
		}
	};
};
