if ("undefined" == typeof(cardbookElementTools)) {
	var { MailServices } = ChromeUtils.import("resource:///modules/MailServices.jsm");
	var { Services } = ChromeUtils.import("resource://gre/modules/Services.jsm");
	var { FormHistory } = ChromeUtils.import("resource://gre/modules/FormHistory.jsm");
	var { cal } = ChromeUtils.import("resource:///modules/calendar/calUtils.jsm");
	var { cardbookRepository } = ChromeUtils.import("chrome://cardbook/content/cardbookRepository.js");

	var cardbookElementTools = {
		
		deleteRowsAllTypes: function () {
			cardbookElementTools.deleteRows('modernRows');
			cardbookElementTools.deleteRows('classicalRows');
		},

		deleteRowsType: function (aType) {
			cardbookElementTools.deleteRows(aType + 'Groupbox');
		},

		deleteRows: function (aObjectName) {
			// for anonid this does not work
			try {
				var aListRows = document.getElementById(aObjectName);
				while (aListRows.hasChildNodes()) {
					aListRows.lastChild.remove();
				}
			} catch (e) {}
		},

		deleteTableRows: function (aTableName, aTableHeaderRowName) {
			let table = document.getElementById(aTableName);
			let toDelete = [];
			for (let row of table.rows) {
				if (aTableHeaderRowName) {
					if (row.id != aTableHeaderRowName) {
						toDelete.push(row);
					}
				} else {
					toDelete.push(row);
				}
			}
			for (let row of toDelete) {
				let oldChild = table.removeChild(row);
			}
		},

		addCategoriesRow: function (aParent, aCategories) {
			cardbookElementTools.deleteRows(aParent.id);

			for (category of aCategories) {
				var aBox = document.createXULElement('box');
				aParent.appendChild(aBox);
				aBox.setAttribute('flex', '1');
				var aLabel = document.createXULElement('label');
				aBox.appendChild(aLabel);
				aLabel.setAttribute('id', category + 'Label');
				aLabel.setAttribute('value', category);
				aLabel.setAttribute('class', 'tagvalue cardbookCategoryClass');
				aLabel.setAttribute('type', 'category_' + cardbookRepository.cardbookUtils.formatCategoryForCss(category));
			}
		},

		addCaption: function (aType, aParent) {
			let aCaption = document.createXULElement('label');
			aParent.appendChild(aCaption);
			aCaption.setAttribute('id', aType + '_caption');
			aCaption.setAttribute('value', cardbookRepository.extension.localeData.localizeMessage(aType + "GroupboxLabel"));
			aCaption.setAttribute('class', 'header');
		},
		
		addCaptionWithLabel: function (aType, aParent, aValue) {
			let aCaption = document.createXULElement('label');
			aParent.appendChild(aCaption);
			aCaption.setAttribute('id', aType + '_caption');
			aCaption.setAttribute('value', aValue);
			aCaption.setAttribute('class', 'header');
		},
		
		addTreeSplitter: function (aParent, aParameters) {
			var aSplitter = document.createXULElement('splitter');
			aSplitter.setAttribute('class', 'tree-splitter');
			for (var prop in aParameters) {
				aSplitter.setAttribute(prop, aParameters[prop]);
			}
			aParent.appendChild(aSplitter);
			return aSplitter;
		},
		
		addHBox: function (aType, aIndex, aParent, aParameters) {
			var aHBox = document.createXULElement('hbox');
			aParent.appendChild(aHBox);
			aHBox.setAttribute('id', aType + '_' + aIndex + '_hbox');
			aHBox.setAttribute('flex', '1');
			aHBox.setAttribute('align', 'center');

			for (var prop in aParameters) {
				aHBox.setAttribute(prop, aParameters[prop]);
			}
			return aHBox;
		},
		
		addVBox: function (aParent, aId, aParameters) {
			var aVBox = document.createXULElement('vbox');
			aParent.appendChild(aVBox);
			aVBox.setAttribute('id', aId);

			for (var prop in aParameters) {
				aVBox.setAttribute(prop, aParameters[prop]);
			}
			return aVBox;
		},
		
		addHTMLElement: function (aElement, aParent, aId, aParameters) {
			let element = document.createElementNS("http://www.w3.org/1999/xhtml", `html:${aElement}`);
			aParent.appendChild(element);
			if (aId) {
				element.setAttribute('id', aId);
			}
			for (let prop in aParameters) {
				element.setAttribute(prop, aParameters[prop]);
			}
			return element;
		},

		addHTMLTABLE: function (aParent, aId, aParameters) {
			let table = cardbookElementTools.addHTMLElement("table", aParent, aId, aParameters);
			return table;
		},

		addHTMLTR: function (aParent, aId, aParameters) {
			let tr = cardbookElementTools.addHTMLElement("tr", aParent, aId, aParameters);
			return tr;
		},

		addHTMLTD: function (aParent, aId, aParameters) {
			let td = cardbookElementTools.addHTMLElement("td", aParent, aId, aParameters);
			return td;
		},

		addHTMLTHEAD: function (aParent, aId, aParameters) {
			let thead = cardbookElementTools.addHTMLElement("thead", aParent, aId, aParameters);
			return thead;
		},

		addHTMLTBODY: function (aParent, aId, aParameters) {
			let tbody = cardbookElementTools.addHTMLElement("tbody", aParent, aId, aParameters);
			return tbody;
		},

		addHTMLTH: function (aParent, aId, aParameters) {
			let th = cardbookElementTools.addHTMLElement("th", aParent, aId, aParameters);
			return th;
		},

		addHTMLIMAGE: function (aParent, aId, aParameters) {
			let image = cardbookElementTools.addHTMLElement("img", aParent, aId, aParameters);
			return image;
		},

		addLabel: function (aOrigBox, aId, aValue, aControl, aParameters) {
			var aLabel = document.createXULElement('label');
			aOrigBox.appendChild(aLabel);
			aLabel.setAttribute('id', aId);
			aLabel.setAttribute('value', aValue);
			aLabel.setAttribute('control', aControl);
			for (var prop in aParameters) {
				aLabel.setAttribute(prop, aParameters[prop]);
			}
			return aLabel;
		},

		addHTMLINPUT: function (aParent, aId, aValue, aParameters) {
			let aTextbox = cardbookElementTools.addHTMLElement("input", aParent, aId, aParameters)
			aTextbox.setAttribute('value', aValue);
			return aTextbox;
		},

		addDatepicker: function (aParent, aId, aValue, aParameters) {
			var aDatepicker = document.createXULElement('cardbookdatepicker');
			aParent.appendChild(aDatepicker);
			aDatepicker.setAttribute('id', aId);
			aDatepicker.setAttribute('value', aValue);

			for (var prop in aParameters) {
				aDatepicker.setAttribute(prop, aParameters[prop]);
			}
			return aDatepicker;
		},

		addHTMLTEXTAREA: function (aParent, aId, aValue, aParameters) {
			let aTextarea = cardbookElementTools.addHTMLElement("textarea", aParent, aId, aParameters)
			aTextarea.value = aValue;
			return aTextarea;
		},

		loadCountries: function (aPopup, aMenu, aDefaultValue, aAddEmptyCountries, aUseCodeValues) {
			var myResult = [];
			for (let code of cardbookRepository.countriesList) {
				let country = cardbookRepository.extension.localeData.localizeMessage("region-name-" + code);
				if (aUseCodeValues) {
					myResult.push([code.toUpperCase(), country]);
				} else {
					myResult.push([country, country]);
				}
			}
			cardbookRepository.cardbookUtils.sortMultipleArrayByString(myResult,1,1);
			cardbookElementTools.deleteRows(aPopup.id);
			var defaultIndex = 0;
			var j = 0;
			if (aAddEmptyCountries) {
				var menuItem = aMenu.appendItem("", "");
				j++;
			}
			var found = false;
			for (var i = 0; i < myResult.length; i++) {
				var menuItem = aMenu.appendItem(myResult[i][1], myResult[i][0]);
				aPopup.appendChild(menuItem);
				if (!found && aDefaultValue != "" && myResult[i][0].toUpperCase() == aDefaultValue.toUpperCase()) {
					defaultIndex=j;
					found=true;
				}
				j++;
			}
			if (found) {
				aMenu.selectedIndex = defaultIndex;
			}
		},

		loadAccountsOrCatsTreeMenu: function (aPopupName, aMenuName, aDefaultId) {
			if (document.getElementById(aMenuName)) {
				var myPopup = document.getElementById(aPopupName);
				cardbookElementTools.deleteRows(aPopupName);
				var defaultIndex = 0;
				var j = 0;
				var typeName = [ 'all', 'enabled', 'disabled', 'local', 'remote', 'search' ];
				for (var i = 0; i < typeName.length; i++) {
					var menuItem = document.getElementById(aMenuName).appendItem(cardbookRepository.extension.localeData.localizeMessage(typeName[i] + "AccountsLabel"), typeName[i]);
					menuItem.setAttribute("type", "radio");
					menuItem.setAttribute("checked", "false");
					myPopup.appendChild(menuItem);
					if (typeName[i] == aDefaultId) {
						defaultIndex=j;
						menuItem.setAttribute("checked", "true");
					}
					j++;
				}
				document.getElementById(aMenuName).selectedIndex = defaultIndex;
			}
		},

		loadGender: function (aPopupName, aMenuName, aDefaultId) {
			var myPopup = document.getElementById(aPopupName);
			cardbookElementTools.deleteRows(aPopupName);
			var defaultIndex = 0;
			var j = 0;
			var myResult = [["", ""]];
			for (var type in cardbookRepository.cardbookGenderLookup) {
				myResult.push([type, cardbookRepository.cardbookGenderLookup[type]]);
			}
			for (var i = 0; i < myResult.length; i++) {
				var menuItem = document.createXULElement("menuitem");
				menuItem.setAttribute("label", myResult[i][1]);
				menuItem.setAttribute("value", myResult[i][0]);
				menuItem.setAttribute("class", "menuitem-iconic");
				menuItem.setAttribute("type", "radio");
				myPopup.appendChild(menuItem);
				if (myResult[i][0] == aDefaultId) {
					defaultIndex=j;
				}
				j++;
			}
			document.getElementById(aMenuName).selectedIndex = defaultIndex;
			document.getElementById(aMenuName).selectedItem.setAttribute("checked", "true");
		},

		loadAddressBooks: function (aPopup, aMenu, aDefaultId, aExclusive, aAddAllABs, aIncludeReadOnly, aIncludeSearch, aIncludeDisabled,
										aInclRestrictionList, aExclRestrictionList) {
			cardbookElementTools.deleteRows(aPopup.id);
			var defaultIndex = 0;
			var j = 0;
			if (aAddAllABs) {
				var menuItem = document.createXULElement("menuitem");
				menuItem.setAttribute("label", cardbookRepository.extension.localeData.localizeMessage("allAddressBooks"));
				menuItem.setAttribute("value", "allAddressBooks");
				menuItem.setAttribute("class", "menuitem-iconic");
				aPopup.appendChild(menuItem);
				if ("allAddressBooks" == aDefaultId) {
					defaultIndex=j;
				}
				j++;
			}
			var sortedAddressBooks = [];
			for (let account of cardbookRepository.cardbookAccounts) {
				if ((aIncludeDisabled || account[2])
						&& (aIncludeReadOnly || !account[4])
						&& (aIncludeSearch || (account[3] !== "SEARCH"))) {
					if (aExclRestrictionList && aExclRestrictionList[account[1]]) {
						continue;
					}
					if (aInclRestrictionList && aInclRestrictionList.length > 0) {
						if (aInclRestrictionList[account[1]]) {
							sortedAddressBooks.push([account[0], account[1], cardbookRepository.getABIconType(account[3])]);
						}
					} else {
						sortedAddressBooks.push([account[0], account[1], cardbookRepository.getABIconType(account[3])]);
					}
				}
			}
			if (!aExclusive) {
				for (let addrbook of MailServices.ab.directories) {
					// remote LDAP directory
					if (addrbook.isRemote && addrbook.dirType === 0) {
						continue;
					}
					if (aInclRestrictionList && aInclRestrictionList.length > 0) {
						if (aInclRestrictionList[addrbook.dirPrefId]) {
							sortedAddressBooks.push([addrbook.dirName, addrbook.dirPrefId, "standard-abook"]);
						}
					} else {
						sortedAddressBooks.push([addrbook.dirName, addrbook.dirPrefId, "standard-abook"]);
					}
				}
			}
			cardbookRepository.cardbookUtils.sortMultipleArrayByString(sortedAddressBooks,0,1);
			for (var i = 0; i < sortedAddressBooks.length; i++) {
				var menuItem = document.createXULElement("menuitem");
				menuItem.setAttribute("label", sortedAddressBooks[i][0]);
				menuItem.setAttribute("value", sortedAddressBooks[i][1]);
				menuItem.setAttribute("ABtype", sortedAddressBooks[i][2]);
				menuItem.setAttribute("class", "menuitem-iconic");
				aPopup.appendChild(menuItem);
				if (sortedAddressBooks[i][1] == aDefaultId) {
					defaultIndex=j;
				}
				j++;
			}
			aMenu.selectedIndex = defaultIndex;
		},

		loadSyncAddressBooks: function (aPopup) {
			cardbookElementTools.deleteRows(aPopup.id);
			var sortedAddressBooks = [];
			for (let account of cardbookRepository.cardbookAccounts) {
				if (account[2] && cardbookRepository.cardbookUtils.isMyAccountRemote(account[3])) {
					sortedAddressBooks.push([account[0], account[1], cardbookRepository.getABIconType(account[3])]);
				}
			}
			cardbookRepository.cardbookUtils.sortMultipleArrayByString(sortedAddressBooks,0,1);
			for (var i = 0; i < sortedAddressBooks.length; i++) {
				var menuItem = document.createXULElement("menuitem");
				menuItem.setAttribute("label", sortedAddressBooks[i][0]);
				menuItem.setAttribute("value", sortedAddressBooks[i][1]);
				menuItem.setAttribute("ABtype", sortedAddressBooks[i][2]);
				menuItem.setAttribute("class", "menuitem-iconic");
				menuItem.addEventListener("command", function(aEvent) {
					cardbookRepository.cardbookSynchronization.syncAccount(this.value);
					aEvent.stopPropagation();
				}, false);
				aPopup.appendChild(menuItem);
			}
		},

		loadCategories: function (aPopupName, aMenuName, aDefaultPrefId, aDefaultCatId, aAddAllCats, aAddOnlyCats, aAddNoCats, aAddEmptyCats, aInclRestrictionList, aExclRestrictionList) {
			var myPopup = document.getElementById(aPopupName);
			cardbookElementTools.deleteRows(aPopupName);
			var defaultIndex = 0;
			var j = 0;
			if (aAddEmptyCats) {
				var menuItem = document.createXULElement("menuitem");
				menuItem.setAttribute("label", "");
				menuItem.setAttribute("value", "");
				menuItem.setAttribute("class", "menuitem-iconic");
				menuItem.setAttribute("type", "radio");
				myPopup.appendChild(menuItem);
				j++;
			}
			if (!(aInclRestrictionList && aInclRestrictionList[aDefaultPrefId])) {
				if (aAddAllCats) {
					var menuItem = document.createXULElement("menuitem");
					menuItem.setAttribute("label", cardbookRepository.extension.localeData.localizeMessage("allCategories"));
					menuItem.setAttribute("value", "allCategories");
					menuItem.setAttribute("class", "menuitem-iconic");
					menuItem.setAttribute("type", "radio");
					myPopup.appendChild(menuItem);
					if ("allCategories" == aDefaultCatId) {
						defaultIndex=j;
					}
					j++;
				}
				if (aAddOnlyCats) {
					var menuItem = document.createXULElement("menuitem");
					menuItem.setAttribute("label", cardbookRepository.extension.localeData.localizeMessage("onlyCategories"));
					menuItem.setAttribute("value", "onlyCategories");
					menuItem.setAttribute("class", "menuitem-iconic");
					menuItem.setAttribute("type", "radio");
					myPopup.appendChild(menuItem);
					if ("onlyCategories" == aDefaultCatId) {
						defaultIndex=j;
					}
					j++;
				}
				if (aAddNoCats) {
					var menuItem = document.createXULElement("menuitem");
					menuItem.setAttribute("label", cardbookRepository.extension.localeData.localizeMessage("noCategory"));
					menuItem.setAttribute("value", "noCategory");
					menuItem.setAttribute("class", "menuitem-iconic");
					menuItem.setAttribute("type", "radio");
					myPopup.appendChild(menuItem);
					if ("noCategory" == aDefaultCatId) {
						defaultIndex=j;
					}
					j++;
				}
			}
			var sortedCategories = [];
			if (cardbookRepository.cardbookAccountsCategories[aDefaultPrefId]) {
				for (let category of cardbookRepository.cardbookAccountsCategories[aDefaultPrefId]) {
					if (aExclRestrictionList && aExclRestrictionList[aDefaultPrefId] && aExclRestrictionList[aDefaultPrefId][category]) {
						continue;
					}
					if (aInclRestrictionList && aInclRestrictionList[aDefaultPrefId]) {
						if (aInclRestrictionList[aDefaultPrefId][category]) {
							sortedCategories.push([category, aDefaultPrefId+"::categories::"+category]);
						}
					} else {
						sortedCategories.push([category, aDefaultPrefId+"::categories::"+category]);
					}
				}
			}
			cardbookRepository.cardbookUtils.sortMultipleArrayByString(sortedCategories,0,1);
			for (var i = 0; i < sortedCategories.length; i++) {
				var menuItem = document.createXULElement("menuitem");
				menuItem.setAttribute("label", sortedCategories[i][0]);
				menuItem.setAttribute("value", sortedCategories[i][1]);
				menuItem.setAttribute("class", "menuitem-iconic");
				menuItem.setAttribute("type", "radio");
				myPopup.appendChild(menuItem);
				if (sortedCategories[i][1] == aDefaultCatId) {
					defaultIndex=j;
				}
				j++;
			}
			document.getElementById(aMenuName).selectedIndex = defaultIndex;
			document.getElementById(aMenuName).selectedItem.setAttribute("checked", "true");
		},

		addPrefStar: function (aParent, aType, aIndex, aValue) {
			var aPrefButton = document.createXULElement('button');
			aParent.appendChild(aPrefButton);
			aPrefButton.setAttribute('id', aType + '_' + aIndex + '_PrefImage');
			aPrefButton.setAttribute('class', 'small-button cardbookPrefStarClass');
			if (aValue) {
				aPrefButton.setAttribute('haspref', 'true');
			} else {
				aPrefButton.removeAttribute('haspref');
			}
			aPrefButton.setAttribute('tooltiptext', cardbookRepository.extension.localeData.localizeMessage("prefLabel"));

			function firePrefCheckBox(event) {
				var myIdArray = this.id.split('_');
				if (document.getElementById(myIdArray[0] + '_' + myIdArray[1] + '_prefWeightBoxLabel') &&
					document.getElementById(myIdArray[0] + '_' + myIdArray[1] + '_prefWeightBox')) {
					var myPrefWeightBoxLabel = document.getElementById(myIdArray[0] + '_' + myIdArray[1] + '_prefWeightBoxLabel');
					var myPrefWeightBox = document.getElementById(myIdArray[0] + '_' + myIdArray[1] + '_prefWeightBox');
					if (this.getAttribute('haspref')) {
						myPrefWeightBoxLabel.disabled = true;
						myPrefWeightBox.disabled = true;
					} else {
						myPrefWeightBoxLabel.disabled = false;
						myPrefWeightBox.disabled = false;
					}
					myPrefWeightBox.value = "";
				}
				if (this.getAttribute('haspref')) {
					this.removeAttribute('haspref');
				} else {
					this.setAttribute('haspref', 'true');
				}
			};
			aPrefButton.addEventListener("command", firePrefCheckBox, false);
			return aPrefButton;
		},

		addKeyButton: function (aParent, aType, aIndex, aKeyType, aCardFn, aCardDirPrefId) {
			var aKeyButton = document.createXULElement('button');
			aParent.appendChild(aKeyButton);
			aKeyButton.setAttribute('id', aType + '_' + aIndex + '_PrefImage');
			aKeyButton.setAttribute('class', 'cardbookKeyClass');
			if (cardbookRepository.cardbookUtils.getPrefBooleanFromTypes(aKeyType.types)) {
				aKeyButton.setAttribute('haspref', 'true');
			} else {
				aKeyButton.removeAttribute('haspref');
			}
			let keyValue = aKeyType.value ? aKeyType.value : aKeyType.URI;
			aKeyButton.setAttribute('keyValue', keyValue);
			aKeyButton.setAttribute('cardFn', aCardFn);
			aKeyButton.setAttribute('cardDirPrefId', aCardDirPrefId);

			function fireKeyCheckBox(event) {
				let confirmTitle = cardbookRepository.extension.localeData.localizeMessage("confirmTitle");
				let confirmMsg = cardbookRepository.extension.localeData.localizeMessage("importKeyFromCards.label");
				if (Services.prompt.confirm(window, confirmTitle, confirmMsg)) {
					wdw_cardbook.importKeyFromValue(
						this.getAttribute('keyValue'),
						{
							'dirPrefId': this.getAttribute('cardDirPrefId'),
							'fn': this.getAttribute('cardFn')
						}
					);
				}
			};
			aKeyButton.addEventListener("command", fireKeyCheckBox, false);
			return aKeyButton;
		}
	};

};
