/**
 * Class used to generate a column picker used for the TreeViewTableHeader in
 * case the visibility of the columns of a table can be changed.
 *
 * Include treeView.ftl for strings.
 */
class CBTreeViewTableColumnPicker extends HTMLTableCellElement {
	/**
	 * The clickable button triggering the picker context menu.
	 * @type {HTMLButtonElement}
	 */
	#button;
	#dropdown;
	#container;

	connectedCallback() {
		if (this.hasConnected) {
			return;
		}
		this.hasConnected = true;

		this.#dropdown = document.createElement("div", {is: "cb-button-dropdown"});
		this.classList.add("tree-table-cell-container", "cb-tree-view-table-column-picker");
		this.#dropdown.classList.add("tree-table-cell-container", "cb-tree-view-table-column-picker", "down");
		this.#button = this.#dropdown.getButton();
		this.#container = this.#dropdown.getContainer();
		this.#container.classList.add("tree-table-cell-container", "cb-tree-view-table-column-picker");
		this.#button.title = messenger.i18n.getMessage("columnPickerTitle");
		this.#button.classList.add("button-flat", "button-column-picker");
		this.appendChild(this.#dropdown);

		const img = document.createElement("img");
		img.src = "";
		img.alt = "";
		this.#button.appendChild(img);

		// this.setAttribute("is", "cb-tree-view-table-column-picker");
		// this.classList.add("tree-table-cell-container");

		// this.#button = document.createElement("button");
		// this.#button.title = messenger.i18n.getMessage("columnPickerTitle");
		// this.#button.classList.add("button-flat", "button-column-picker");
		// this.appendChild(this.#button);

		// const img = document.createElement("img");
		// img.src = "";
		// img.alt = "";
		// this.#button.appendChild(img);

		let treeview = this.closest("tree-view");
		let table = this.closest("table");
		let columns = table.columns;
		columns.sort(function(a, b){ return a.label.localeCompare(b.label) })
		columns.sort(function(a, b){ return a.ordinal-b.ordinal })
		let options = [];
		for (let column of columns) {
			let checked = !column.hidden;
			options.push([`${column.label}`, `${treeview.id}-Picker-${column.id}`, checked]);
		}

		function columnsChanged(event) {
			let value = event.target.id.replace(`${treeview.id}-Picker-`, "");
			table.dispatchEvent(new CustomEvent("columns-changed", {bubbles: true, detail: {target: event.target, value: value}}));
		}
		function restoreColumns(event) {
			table.dispatchEvent(new CustomEvent("restore-columns", {bubbles: true}));
			event.stopImmediatePropagation();
		}
		let restId = `${treeview.id}-Picker-RestoreDefault`;
		let restLabel = messenger.i18n.getMessage("columnPickerRestoreColumnOrder");
		
		let anchor = document.getElementById(`${treeview.id}Picker`);
		this.#dropdown.loadOptions(`${treeview.id}-Picker-RestoreDefault`, anchor, options, columnsChanged, restId, restLabel, restoreColumns);
		this.#dropdown.setPosition("down");
		this.#dropdown.setPosition(treeview.id == "availableCardsTree" ? "left" : "right");
	}
}
customElements.define("cb-tree-view-table-column-picker", CBTreeViewTableColumnPicker,{ extends: "th" });
