import { cardbookHTMLDates } from "./cardbookHTMLDates.mjs";
import { cardbookNewPreferences } from "./preferences/cardbookNewPreferences.mjs";
import { cardbookHTMLFormulas } from "./cardbookHTMLFormulas.mjs";
import { cardbookHTMLTypes } from "./cardbookHTMLTypes.mjs";

export var cardbookHTMLUtils = {

	multilineFields: [ "email", "tel", "adr", "impp", "url" ],
	newFields: [ "gender", "birthplace", "anniversary", "deathdate", "deathplace" ],
	adrElements: [ "postOffice", "extendedAddr", "street", "locality", "region", "postalCode", "country" ],
	defaultEmailFormat: "X-MOZILLA-HTML",
	prefCSVPrefix: "*:",
	countriesList: ["ad","ae","af","ag","ai","al","am","ao","aq","ar","as","at","au","aw","az","ba","bb","bd","be","bf","bg",
						"bh","bi","bj","bl","bm","bn","bo","bq","br","bs","bt","bv","bw","by","bz","ca","cc","cd","cf","cg","ch",
						"ci","ck","cl","cm","cn","co","cp","cr","cu","cv","cw","cx","cy","cz","de","dg","dj","dk","dm","do","dz",
						"ec","ee","eg","eh","er","es","et","fi","fj","fk","fm","fo","fr","ga","gb","gd","ge","gf","gg","gh","gi",
						"gl","gm","gn","gp","gq","gr","gs","gt","gu","gw","gy","hk","hm","hn","hr","ht","hu","id","ie","il","im",
						"in","io","iq","ir","is","it","je","jm","jo","jp","ke","kg","kh","ki","km","kn","kp","kr","kw","ky","kz",
						"la","lb","lc","li","lk","lr","ls","lt","lu","lv","ly","ma","mc","md","me","mf","mg","mh","mk","ml","mm",
						"mn","mo","mp","mq","mr","ms","mt","mu","mv","mw","mx","my","mz","na","nc","ne","nf","ng","ni","nl","no",
						"np","nr","nu","nz","om","pa","pe","pf","pg","ph","pk","pl","pm","pn","pr","pt","pw","py","qa","qm","qs",
						"qu","qw","qx","qz","re","ro","rs","ru","rw","sa","sb","sc","sd","se","sg","sh","si","sk","sl","sm","sn",
						"so","sr","ss","st","sv","sx","sy","sz","tc","td","tf","tg","th","tj","tk","tl","tm","tn","to","tr","tt",
						"tv","tw","tz","ua","ug","us","uy","uz","va","vc","ve","vg","vi","vn","vu","wf","ws","xa","xb","xc","xd",
						"xe","xg","xh","xj","xk","xl","xm","xp","xq","xr","xs","xt","xu","xv","xw","ye","yt","za","zm","zw"],

	sortArrayByString: function (aArray, aInvert) {
		function compare(a, b) { return a.localeCompare(b)*aInvert; };
		return aArray.sort(compare);
	},

	sortMultipleArrayByString: function (aArray, aIndex, aInvert) {
		function compare(a, b) { return a[aIndex].localeCompare(b[aIndex])*aInvert; };
		return aArray.sort(compare);
	},

	sortMultipleArrayByNumber: function (aArray, aIndex, aInvert) {
		function compare(a, b) { return (a[aIndex] - b[aIndex])*aInvert; };
		return aArray.sort(compare);
	},

	arrayUnique: function (array) {
		var a = array.concat();
		for (var i=0; i<a.length; ++i) {
			for (var j=i+1; j<a.length; ++j) {
				if (a[i] == a[j])
					a.splice(j--, 1);
			}
		}
		return a;
	},

	cleanArray: function (vArray) {
		var newArray = [];
		for(let i = 0; i<vArray.length; i++){
			if (vArray[i] && vArray[i] != ""){
				newArray.push(vArray[i].trim());
			}
		}
		return newArray;
	},

	cleanArrayWithoutTrim: function (vArray) {
		var newArray = [];
		for(let i = 0; i<vArray.length; i++){
			if (vArray[i] && vArray[i] != ""){
				newArray.push(vArray[i]);
			}
		}
		return newArray;
	},
	
	escapeString: function (vString) {
		return vString.replace(/\\;/g,"@ESCAPEDSEMICOLON@").replace(/\\,/g,"@ESCAPEDCOMMA@");
	},

	escapeString1: function (vString) {
		return vString.replace(/\\\(/g,"@ESCAPEDLEFTPARENTHESIS@").replace(/\\\)/g,"@ESCAPEDRIGHTPARENTHESIS@").replace(/\\\|/g,"@ESCAPEDPIPE@"); 
	},

	escapeArray2: function (vArray) {
		var result = []
		for (let i = 0; i<vArray.length; i++){
			if (vArray[i] && vArray[i] != ""){
				result[i] = vArray[i].replace(/\(/g,"@ESCAPEDLEFTPARENTHESIS@").replace(/\)/g,"@ESCAPEDRIGHTPARENTHESIS@").replace(/\|/g,"@ESCAPEDPIPE@");
			} else {
				result[i] = "";
			}
		}
		return result;
	},

	escapeArray: function (vArray) {
		for (let i = 0; i<vArray.length; i++){
			if (vArray[i] && vArray[i] != ""){
				vArray[i] = vArray[i].replace(/\\;/g,"@ESCAPEDSEMICOLON@").replace(/\\,/g,"@ESCAPEDCOMMA@");
			}
		}
		return vArray;
	},

	escapeStringSemiColon: function (vString) {
		return vString.replace(/;/g,"@ESCAPEDSEMICOLON@");
	},

	unescapeStringSemiColon: function (vString) {
		return vString.replace(/@ESCAPEDSEMICOLON@/g,"\\;");
	},

	unescapeString: function (vString) {
		return vString.replace(/@ESCAPEDSEMICOLON@/g,";").replace(/\\;/g,";").replace(/@ESCAPEDCOMMA@/g,",").replace(/\\,/g,",");
	},

	unescapeString1: function (vString) {
		return vString.replace(/@ESCAPEDLEFTPARENTHESIS@/g,"(").replace(/@ESCAPEDRIGHTPARENTHESIS@/g,")").replace(/@ESCAPEDPIPE@/g,"|");
	},

	unescapeArray: function (vArray) {
		for (let i = 0; i<vArray.length; i++){
			if (vArray[i] && vArray[i] != ""){
				vArray[i] = cardbookHTMLUtils.unescapeString(vArray[i]);
			}
		}
		return vArray;
	},

	getRadioNodes: function (aName) {
		let searchString = `input[type='radio'][name='${aName}']`;
		return document.querySelectorAll(searchString);
	},

	getRadioValue: function (aName) {
		for (let node of this.getRadioNodes(aName)) {
			if (node.checked) {
				return node.value;
			}
		}
		return null;
	},

	getPrefValue: function (aName) {
		return cardbookNewPreferences.getPref(aName);
	},

	setPrefValue: async function (aName, aValue) {
		await cardbookNewPreferences.setPref(aName, aValue);
		await messenger.runtime.sendMessage({query: "cardbook.pref.setLegacyPref", key: aName, value: aValue});
	},

	delBranch: async function (aStartingPoint) {
		let keys = cardbookNewPreferences.getBranchKeys(aStartingPoint);
		await cardbookNewPreferences.removePrefs(keys);
		await messenger.runtime.sendMessage({query: "cardbook.pref.removeLegacyPrefs", keys: keys})
	},

	randomChannel: function(brightness) {
		var r = 255-brightness;
		var n = 0|((Math.random() * r) + brightness);
		var s = n.toString(16);
		return (s.length==1) ? "0"+s : s;
	},

	randomColor: function(brightness) {
		return "#" + this.randomChannel(brightness) + this.randomChannel(brightness) + this.randomChannel(brightness);
	},

	convertField: function(aFunction, aValue) {
		switch(aFunction) {
			case "lowercase":
				return aValue.toLowerCase();
				break;
			case "uppercase":
				return aValue.toUpperCase();
				break;
			case "capitalization":
				return aValue.charAt(0).toUpperCase() + aValue.substr(1).toLowerCase();
				break;
		};
	},

	getTextColorFromBackgroundColor: function (aHexBackgroundColor) {
		function hexToRgb(aColor) {
			var result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(aColor);
			return result ? {r: parseInt(result[1], 16), g: parseInt(result[2], 16), b: parseInt(result[3], 16)} : null;
		}

		var rgbColor = hexToRgb(aHexBackgroundColor);
		// http://www.w3.org/TR/AERT#color-contrast
		var o = Math.round(((parseInt(rgbColor.r) * 299) + (parseInt(rgbColor.g) * 587) + (parseInt(rgbColor.b) * 114)) / 1000);
		var fore = (o > 125) ? "black" : "white";
		return fore;
	},

	getTime: function() {
		var objToday = new Date();
		var year = objToday.getFullYear();
		var month = ("0" + (objToday.getMonth() + 1)).slice(-2);
		var day = ("0" + objToday.getDate()).slice(-2);
		var hour = ("0" + objToday.getHours()).slice(-2);
		var min = ("0" + objToday.getMinutes()).slice(-2);
		var sec = ("0" + objToday.getSeconds()).slice(-2);
		var msec = ("00" + objToday.getMilliseconds()).slice(-3);
		return year + "." + month + "." + day + " " + hour + ":" + min + ":" + sec + ":" + msec;
	},

	deleteCssAllRules: function (aStyleSheet) {
		// aStyleSheet.cssRules may not be available
		try {
			while(aStyleSheet.cssRules.length > 0) {
				aStyleSheet.deleteRule(0);
			}
		} catch(e) {}

	},

	createMarkerRule: function (aStyleSheet, aStyleSheetRuleName) {
		var ruleString = "." + aStyleSheetRuleName + " {}";
		var ruleIndex = aStyleSheet.insertRule(ruleString, aStyleSheet.cssRules.length);
	},

	createCssCardRulesForTable: function (aStyleSheet, aDirPrefId, aColor, aUseColor) {
		let oppositeColor = cardbookHTMLUtils.getTextColorFromBackgroundColor(aColor);
		if (aUseColor == "text") {
			let ruleString1 = ".tableTree tr[data-value='SEARCH color_" + aDirPrefId + "']:not([rowSelected='true']) {color: " + aColor + " !important;}";
			let ruleIndex1 = aStyleSheet.insertRule(ruleString1, aStyleSheet.cssRules.length);

			let ruleString2 = ".tableTree tr[data-value='SEARCH color_" + aDirPrefId + "']:not([rowSelected='true']) {background-color: " + oppositeColor + " !important;}";
			let ruleIndex2 = aStyleSheet.insertRule(ruleString2, aStyleSheet.cssRules.length);
		} else if (aUseColor == "background") {
			let ruleString1 = ".tableTree tr[data-value='SEARCH color_" + aDirPrefId + "']:not([rowSelected='true']) {background-color: " + aColor + " !important;}";
			let ruleIndex1 = aStyleSheet.insertRule(ruleString1, aStyleSheet.cssRules.length);

			let ruleString2 = ".tableTree tr[data-value='SEARCH color_" + aDirPrefId + "']:not([rowSelected='true']) {color: " + oppositeColor + " !important;}";
			let ruleIndex2 = aStyleSheet.insertRule(ruleString2, aStyleSheet.cssRules.length);
		}
	},

	getDateFormat: function (aDirPrefId, aVersion) {
		let type = cardbookNewPreferences.getType(aDirPrefId);
		if ( type == "APPLE" || type == "YAHOO") {
			return "YYYY-MM-DD";
		} else {
			return aVersion;
		}
	},

	blobToBase64: function (blob) {
		return new Promise((resolve, _) => {
			const reader = new FileReader();
			reader.onloadend = () => resolve(reader.result);
			reader.readAsDataURL(blob);
		});
	},

	cleanCategories: function (aCategoryList) {
		let uncat = cardbookHTMLUtils.getPrefValue("uncategorizedCards");
		function filterCategories(element) {
			return (element != uncat);
		}
		return cardbookHTMLUtils.arrayUnique(aCategoryList.filter(filterCategories));
	},

	formatCategoryForCss: function (aCategory) {
		return aCategory.replace(/[!\"#$%&'\(\)\*\+,\.\/:;<=>\?\@\[\\\]\^`\{\|\}~ ]/g, "_");
	},

	// 2 : HTML, 1 : PlainText, 0 : Unknown
	getMailFormatFromCard: function (aCard) {
		if (aCard) {
			for (let other of aCard.others) {
				let localDelim1 = other.indexOf(":",0);
				if (localDelim1 >= 0) {
					let header = other.substr(0,localDelim1).toUpperCase();
					let trailer = other.substr(localDelim1+1, other.length).toUpperCase();
					if (header == cardbookHTMLUtils.defaultEmailFormat) {
						if (trailer == "TRUE") {
							return "2";
						} else if (trailer == "FALSE") {
							return "1";
						}
						break;
					}
				}
			}
		}
		return "0";
	},

	addEventstoCard: function(aCard, aEventsArray, aPGNextNumber, aDateFormat) {
		var myEventsArray = [];
		for (var i = 0; i < aEventsArray.length; i++) {
			var myValue = cardbookHTMLDates.getVCardDateFromDateString(aEventsArray[i][0], aDateFormat);
			if (aEventsArray[i][2]) {
				myEventsArray.push("ITEM" + aPGNextNumber + ".X-ABDATE;TYPE=PREF:" + myValue);
			} else {
				myEventsArray.push("ITEM" + aPGNextNumber + ".X-ABDATE:" + myValue);
			}
			myEventsArray.push("ITEM" + aPGNextNumber + ".X-ABLABEL:" + aEventsArray[i][1]);
			aPGNextNumber++;
		}
		aCard.others = myEventsArray.concat(aCard.others);
	},

	getEventsFromCard: function(aCardNoteArray, aCardOthers) {
		var myResult = [];
		var myRemainingNote = [];
		var myRemainingOthers = [];
		var eventInNoteEventPrefix = messenger.i18n.getMessage("eventInNoteEventPrefix");
		var typesList = [ "Birthday" , eventInNoteEventPrefix ];
		for (var i = 0; i < aCardNoteArray.length; i++) {
			var found = false;
			for (var j in typesList) {
				var myType = typesList[j];
				// compatibility when not localized
				var EmptyParamRegExp1 = new RegExp("^" + myType + ":([^:]*):(.*)", "ig");
				if (aCardNoteArray[i].replace(EmptyParamRegExp1, "$1") != aCardNoteArray[i]) {
					var lNotesName = aCardNoteArray[i].replace(EmptyParamRegExp1, "$1").replace(/^\s+|\s+$/g,"");
					if (aCardNoteArray[i].replace(EmptyParamRegExp1, "$2") != aCardNoteArray[i]) {
						var lNotesDateFound = aCardNoteArray[i].replace(EmptyParamRegExp1, "$2").replace(/^\s+|\s+$/g,"");
						if (lNotesDateFound.endsWith(":PREF")) {
							myResult.push([lNotesDateFound.replace(":PREF", ""), lNotesName, true]);
						} else {
							myResult.push([lNotesDateFound, lNotesName, false]);
						}
						found = true;
						break;
					}
				}
			}
			if (!found) {
				myRemainingNote.push(aCardNoteArray[i]);
			}
		}
		while (myRemainingNote[0] == "") {
			myRemainingNote.shift();
		}
		// should parse this, may in wrong order, maybe in lowercase
		// ITEM1.X-ABDATE;TYPE=PREF:20200910
		// ITEM1.X-ABLABEL:fiesta
		var myPGToBeParsed = {};
		for (var i = 0; i < aCardOthers.length; i++) {
			var relative = []
			relative = aCardOthers[i].match(/^ITEM([0-9]*)\.(.*)\:(.*)/i);
			if (relative && relative[1] && relative[2] && relative[3]) {
				var myPGName = "ITEM" + relative[1];
				var relativeKey = relative[2].match(/^([^\;]*)/i);
				var key = relativeKey[1].toUpperCase();
				if (!myPGToBeParsed[myPGName]) {
					myPGToBeParsed[myPGName] = ["", "", "", ""];
				}
				if (relative[2].toUpperCase().startsWith("X-ABLABEL")) {
					myPGToBeParsed[myPGName][1] = relative[3];
				} else {
					myPGToBeParsed[myPGName][0] = relative[3];
					myPGToBeParsed[myPGName][2] = relative[2].replace(key, "");
					myPGToBeParsed[myPGName][3] = key;
				}
			} else {
				myRemainingOthers.push(aCardOthers[i]);
			}
		}
		for (var i in myPGToBeParsed) {
			if (myPGToBeParsed[i][3] == "X-ABDATE") {
				myResult.push([myPGToBeParsed[i][0], myPGToBeParsed[i][1], myPGToBeParsed[i][2]]);
			} else {
				if (myPGToBeParsed[i][2]) {
					myRemainingOthers.push(i + "." + myPGToBeParsed[i][3] + myPGToBeParsed[i][2] + ":" + myPGToBeParsed[i][0]);
				} else {
					myRemainingOthers.push(i + "." + myPGToBeParsed[i][3] + ":" + myPGToBeParsed[i][0]);
				}
				myRemainingOthers.push(i + ".X-ABLABEL:" + myPGToBeParsed[i][1]);
			}
		}
		// to change pref for myResult
		return {result: myResult, remainingNote: myRemainingNote, remainingOthers: myRemainingOthers};
	},

	rebuildAllPGs: function (aCard) {
		let myPgNumber = 1;
		for (let field of cardbookHTMLUtils.multilineFields) {
			for (var j = 0; j < aCard[field].length; j++) {
				let myTempString = aCard[field][j][2];
				if (myTempString.startsWith("ITEM")) {
					aCard[field][j][2] = "ITEM" + myPgNumber;
					myPgNumber++;
				}
			}
		}
		let myNewOthers = [];
		let myPGMap = {};
		for (var j = 0; j < aCard.others.length; j++) {
			let myTempString = aCard.others[j];
			var relative = []
			relative = myTempString.match(/^ITEM([0-9]*)\.(.*)/i);
			if (relative && relative[1] && relative[2]) {
				if (myPGMap[relative[1]]) {
					myNewOthers.push("ITEM" + myPGMap[relative[1]] + "." + relative[2]);
				} else {
					myNewOthers.push("ITEM" + myPgNumber + "." + relative[2]);
					myPGMap[relative[1]] = myPgNumber;
					myPgNumber++;
				}
			} else {
				myNewOthers.push(aCard.others[j]);
			}
		}
		aCard.others = JSON.parse(JSON.stringify(myNewOthers));
		return myPgNumber;
	},

	getABTypeFormat: function (aType) {
		switch(aType) {
			case "DIRECTORY":
			case "FILE":
			case "LOCALDB":
			case "CARDDAV":
			case "SEARCH":
				return "CARDDAV";
				break;
		};
		return aType;
	},

	getOnlyTypesFromTypes: function(aArray) {
		function deletePrefs(element) {
			return !(element.toUpperCase().replace(/TYPE=PREF/i,"PREF").replace(/PREF=[0-9]*/i,"PREF") == "PREF");
		}
		let result = [];
		for (let type of aArray) {
			let upperElement = type.toUpperCase();
			if (upperElement == "PREF" || upperElement == "TYPE=PREF") {
				continue;
			} else if (upperElement == "HOME" || upperElement == "FAX" || upperElement == "CELL" || upperElement == "WORK") {
				result.push(type);
			} else if (upperElement.replace(/^TYPE=/i,"") !== upperElement) {
				let tmpArray = type.replace(/^TYPE=/ig,"").split(",").filter(deletePrefs);
				for (let type of tmpArray) {
					if (type == "VOICE" || type == "INTERNET") {
						continue;
					}
					result.push(type);
				}
			}
		}
		return result;
	},

	getNotTypesFromTypes: function(aArray) {
		let result = [];
		for (let element of aArray) {
			var upperElement = element.toUpperCase();
			if (upperElement === "PREF" || upperElement === "TYPE=PREF") {
				continue;
			} else if (upperElement === "HOME" || upperElement === "FAX" || upperElement === "CELL" || upperElement === "WORK") {
				continue;
			} else if (upperElement.replace(/PREF=[0-9]*/i,"PREF") == "PREF") {
				continue;
			} else if (upperElement.replace(/^TYPE=/i,"") === upperElement) {
				result.push(element);
			}
		}
		return result.join(",");
	},

	getPrefBooleanFromTypes: function(aArray) {
		if (aArray) {
			for (let type of aArray) {
				var upperElement = type.toUpperCase();
				if (upperElement === "PREF" || upperElement === "TYPE=PREF") {
					return true;
				} else if (upperElement.replace(/PREF=[0-9]*/i,"PREF") == "PREF") {
					return true;
				} else if (upperElement.replace(/^TYPE=/ig,"") !== upperElement) {
					var tmpArray = type.replace(/^TYPE=/ig,"").split(",");
					for (var j = 0; j < tmpArray.length; j++) {
						var upperElement1 = tmpArray[j].toUpperCase();
						if (upperElement1 === "PREF") {
							return true;
						} else if (upperElement1.replace(/PREF=[0-9]*/i,"PREF") == "PREF") {
							return true;
						}
					}
				}
			}
		}
		return false;
	},

	formatExtension: function (aExtension, aVersion) {
		switch (aExtension) {
			case "pgp":
			case "pub":
				aExtension = "pgp";
				break;
			case "JPG":
			case "jpg":
				aExtension = "jpeg";
				break;
			case "TIF":
			case "tif":
				aExtension = "tiff";
				break;
			case "":
				aExtension = "jpeg";
		}
		if (aVersion == "4.0") {
			aExtension = aExtension.toLowerCase();
		} else {
			aExtension = aExtension.toUpperCase();
		}
		return aExtension;
	},

	formatTelForSearching: function (aString) {
		// +33 6 45 44 42 25 should be equal to 06 45 44 42 25 should be equal to 00 33 6 45 44 42 25 should be equal to 0645444225
		return aString.replace(/^\+\d+\s+/g, "").replace(/^00\s+\d+\s+/g, "").replace(/\D/g, "").replace(/^0/g, "");
	},

	formatTypesForDisplay: function (aTypeList) {
		aTypeList = cardbookHTMLUtils.cleanArray(aTypeList);
		return cardbookHTMLUtils.sortArrayByString(aTypeList, 1).join("    ");
	},

	getFileExtension: function (aFile) {
		var myFileArray = aFile.split("/");
		var myFileArray1 = myFileArray[myFileArray.length-1].split("\\");
		return cardbookHTMLUtils.getFileNameExtension(myFileArray1[myFileArray1.length-1]);
	},

	getFileNameExtension: function (aFileName) {
		let myFileArray = aFileName.split(".");
		let myExtension = "";
		if (myFileArray.length != 1) {
			myExtension = myFileArray[myFileArray.length - 1];
		}
		return myExtension;
	},

	getName: function (aCard) {
		let showNameAs = cardbookHTMLUtils.getPrefValue("showNameAs");
		if (aCard.isAList || showNameAs == "DSP") {
			return aCard.fn;
		}
		if (aCard.lastname != "" && aCard.firstname != "") {
			let result = "";
			if (showNameAs == "LF") {
				result = aCard.lastname + " " + aCard.firstname;
			} else if (showNameAs == "FL") {
				result = aCard.firstname + " " + aCard.lastname;
			} else if (showNameAs == "LFCOMMA") {
				result = aCard.lastname + ", " + aCard.firstname;
			}
			return result.trim();
		} else {
			return aCard.fn;
		}
	},

	convertVCard: async function (aCard, aTargetName, aTargetVersion, aDateFormatSource, aDateFormatTarget) {
		let converted = false;
		let possibleCustomFields = await messenger.runtime.sendMessage({query: "cardbook.getPossibleCustomFields"});
		// basic fields
		if (aCard.version != aTargetVersion) {
			converted = true;
			aCard.version = aTargetVersion;
			for (let newField of cardbookHTMLUtils.newFields) {
				let oldField = "X-" + newField.toUpperCase();
				if (aTargetVersion == "3.0") {
					if (aCard[newField] != "") {
						aCard.others.push(oldField + ":" + aCard[newField]);
						if (!possibleCustomFields[oldField].add && !possibleCustomFields[oldField].added) {
							possibleCustomFields[oldField].add = true;
						}
						aCard[newField] = "";
					}
				} else if (aTargetVersion == "4.0") {
					for (let k = 0; k < aCard.others.length; k++) {
						if (aCard.others[k].startsWith(oldField + ":")) {
							let newFieldRegExp = new RegExp("^" + oldField + ":");
							aCard[newField] = aCard.others[k].replace(newFieldRegExp, "");
							aCard.others.splice(k,1);
							break;
						}
					}
				}
			}

			// lists
			let kindCustom = cardbookHTMLUtils.getPrefValue("kindCustom");
			let memberCustom = cardbookHTMLUtils.getPrefValue("memberCustom");
			if (aCard.isAList) {
				if (aTargetVersion == "3.0") {
					cardbookHTMLUtils.addMemberstoCard(aCard, aCard.member, aCard.kind);
					aCard.member = "";
					aCard.kind = "";
				} else if (aTargetVersion == "4.0") {
					let myMembers = [];
					let myGroup = "";
					for (let j = 0; j < aCard.others.length; j++) {
						if (aCard.others[j].startsWith(memberCustom + ":")) {
							let myFieldRegExp = new RegExp("^" + memberCustom + ":");
							myMembers.push(aCard.others[j].replace(myFieldRegExp, ""));
							aCard.others.splice(j,1);
							j--;
						} else if (aCard.others[j].startsWith(kindCustom + ":")) {
							let myFieldRegExp = new RegExp("^" + kindCustom + ":");
							myGroup = aCard.others[j].replace(myFieldRegExp, "");
							aCard.others.splice(j,1);
							j--;
						}
					}
					cardbookHTMLUtils.addMemberstoCard(aCard, myMembers, myGroup);
				}
			}
		}
		// date fields
		if (aDateFormatSource != aDateFormatTarget) {
			if (await cardbookHTMLDates.convertCardDate(aCard, aTargetName, aDateFormatSource, aDateFormatTarget)) {
				converted = true;
			}
		}
		return converted;
	},

	addMemberstoCard: function(aCard, aMemberLines, aKindValue) {
		if (aCard.version == "4.0") {
			aCard.member = JSON.parse(JSON.stringify(aMemberLines));
			if (aKindValue) {
				aCard.kind = aKindValue;
			} else {
				aCard.kind = "group";
			}
		} else if (aCard.version == "3.0") {
			let kindCustom = cardbookHTMLUtils.getPrefValue("kindCustom");
			let memberCustom = cardbookHTMLUtils.getPrefValue("memberCustom");
			for (var i = 0; i < aCard.others.length; i++) {
				let localDelim1 = aCard.others[i].indexOf(":",0);
				if (localDelim1 >= 0) {
					var header = aCard.others[i].substr(0, localDelim1).toUpperCase();
					var trailer = aCard.others[i].substr(localDelim1+1, aCard.others[i].length);
					if (header == kindCustom || header == memberCustom) {
						aCard.others.splice(i, 1);
						i--;
						continue;
					}
				}
			}
			for (var i = 0; i < aMemberLines.length; i++) {
				if (i === 0) {
					if (aKindValue) {
						aCard.others.push(kindCustom + ":" + aKindValue);
					} else {
						aCard.others.push(kindCustom + ":group");
					}
				}
				aCard.others.push(memberCustom + ":" + aMemberLines[i]);
			}
		}
	},

	formatAddress: function(aAddress, aAdrFormula) {
		let result =  "";
		var adrFormula = cardbookHTMLUtils.getPrefValue("adrFormula");
		let defaultAdrFormula = messenger.i18n.getMessage("addressFormatFormula");
		let myAdrFormula = aAdrFormula ||
							adrFormula ||
							defaultAdrFormula;
		let addressArray = JSON.parse(JSON.stringify(aAddress));
		// maybe a country code (Google uses them)
		let lcRegionCode = addressArray[6].toLowerCase();
		if (cardbookHTMLUtils.countriesList.includes(lcRegionCode)) {
			addressArray[6] = messenger.i18n.getMessage("region-name-" + lcRegionCode);
		}
		result = cardbookHTMLFormulas.getStringFromFormula(myAdrFormula, addressArray);
		var re = /[\n\u0085\u2028\u2029]|\r\n?/;
		var myAdrResultArray = result.split(re);
		return cardbookHTMLUtils.cleanArray(myAdrResultArray).join("\n");
	},

	getCardRegion: function (aCard) {
		let i = 0;
		while (true) {
			if (aCard.adr[i] && aCard.adr[i][0]) {
				var country = aCard.adr[i][0][6].toUpperCase();
				if (country != "") {
					let lcRegionCode = country.toLowerCase();
					// maybe a country code
					if (cardbookHTMLUtils.countriesList.includes(lcRegionCode)) {
						return country;
					}
					// let's try to find a known country
					for (let code of cardbookHTMLUtils.countriesList) {
						let locName = messenger.i18n.getMessage("region-name-" + code);
						if (lcRegionCode == locName.toLowerCase()) {
							return code.toUpperCase();
						}
					}
				}
				i++;
			} else {
				return cardbookHTMLUtils.getPrefValue("defaultRegion");
			}
		}
		return "";
	},

	openURL: async function (aURL) {
		let localizeTarget = cardbookHTMLUtils.getPrefValue("localizeTarget");
		if (localizeTarget == "in") {
			await messenger.runtime.sendMessage({query: "cardbook.openTab", url: aURL});
		} else {
			await messenger.runtime.sendMessage({query: "cardbook.openExternalURL", link: aURL})
		}
	},

	getAddressesFromCards: function (aListOfCards) {
		let listOfAddresses= [];
		if (aListOfCards) {
			for (let card of aListOfCards) {
				for (let adr of card.adr) {
					let adress = adr[0];
					listOfAddresses.push(adress);
				}
			}
		}
		return listOfAddresses;
	},

	localizeCards: function (aListOfSelectedCard, aListOfSelectedAddresses) {
		let listOfAddresses = [];
		if (aListOfSelectedCard) {
			listOfAddresses = cardbookHTMLUtils.getAddressesFromCards(aListOfSelectedCard);
		} else if (aListOfSelectedAddresses) {
			listOfAddresses = JSON.parse(JSON.stringify(aListOfSelectedAddresses));
		}
		
		let localizeEngine = cardbookHTMLUtils.getPrefValue("localizeEngine");
		let urlEngine = "";
		if (localizeEngine === "GoogleMaps") {
			urlEngine = "https://www.google.com/maps?q=";
		} else if (localizeEngine === "OpenStreetMap") {
			urlEngine = "https://www.openstreetmap.org/search?query=";
		} else if (localizeEngine === "BingMaps") {
			urlEngine = "https://www.bing.com/maps/?q=";
		} else {
			return;
		}

		function getAddressURL(address) {
			let result = JSON.parse(JSON.stringify(address));
			result.shift();
			result = result.join("+").replace(/[\n\u0085\u2028\u2029]|\r\n?/g, "+").replace(/ /g, "+");
			return result;
		}

		if (localizeEngine === "GoogleMaps") {
			let url = urlEngine;
			for (let address of listOfAddresses) {
				url = url + getAddressURL(address) + "/";
			}
			cardbookHTMLUtils.openURL(url);
		} else {
			for (let address of listOfAddresses) {
				let url = urlEngine + getAddressURL(address);
				cardbookHTMLUtils.openURL(url);
			}
		}
	},

	// not possible to include prefs for ADR field
	getCardValueByField: function(aCard, aField, aIncludePref) {
		var result = [];
		if (aField.indexOf("_") > 0) {
			var myFieldArray = aField.split("_");
			var myField = myFieldArray[0];
			var myPosition = myFieldArray[1];
			if (myField == "org") {
				var orgStructure = cardbookHTMLUtils.getPrefValue("orgStructure");
				if (orgStructure != "") {
					var myOrgValue = cardbookHTMLUtils.unescapeArray(cardbookHTMLUtils.escapeString(aCard[myField]).split(";"));
					result.push(myOrgValue[myPosition]);
				} else {
					result.push(aCard[myField]);
				}
			// adr_all, adr_5_all, adr_adr_privé, adr_5_adr_privé
			} else if (myField == "adr") {
				var myFieldArray = aField.split("_");
				var myField = myFieldArray[0];
				if (aCard[myField].length) {
					if (["0", "1", "2", "3", "4", "5", "6"].includes(myFieldArray[1])) {
						let subField = parseInt(myFieldArray[1]);
						let type = myFieldArray.slice(2).join("_");
						if (type == "all") {
							if (aCard[myField]) {
								for (var i = 0; i < aCard[myField].length; i++) {
									if (aCard[myField][i][0][subField]) {
										result.push(aCard[myField][i][0][subField]);
									}
								}
							}
						} else if (type == "notype") {
							if (aCard[myField]) {
								for (var i = 0; i < aCard[myField].length; i++) {
									let inputTypes = cardbookHTMLUtils.getOnlyTypesFromTypes(aCard[myField][i][1]);
									let emptyType = inputTypes.length == 0 || inputTypes[0] == "";
									if (emptyType && aCard[myField][i][3].length == 0 && aCard[myField][i][2] == "") {
										if (aCard[myField][i][0][subField]) {
											result.push(aCard[myField][i][0][subField]);
										}
									}
								}
							}
						} else {
							for (var i = 0; i < aCard[myField].length; i++) {
								let found = false;
								if (aCard[myField][i][0][subField]) {
									if (aCard[myField][i][3].length != 0 && aCard[myField][i][2] != "") {
										for (var j = 0; j < aCard[myField][i][3].length; j++) {
											let tmpArray = aCard[myField][i][3][j].split(":");
											if (tmpArray[0] == "X-ABLABEL" && tmpArray[1].replaceAll(" ", "_") == type) {
												found = true;
												break;
											}
										}
									}
									if (found) {
										if (aIncludePref && cardbookHTMLUtils.getPrefBooleanFromTypes(aCard[myField][i][1])) {
											result.push(cardbookHTMLUtils.prefCSVPrefix + aCard[myField][i][0][subField]);
										} else {
											result.push(aCard[myField][i][0][subField]);
										}
									} else {
										var ABType = cardbookNewPreferences.getType(aCard.dirPrefId);
										var ABTypeFormat = cardbookHTMLUtils.getABTypeFormat(ABType);
										var myInputTypes = cardbookHTMLUtils.getOnlyTypesFromTypes(aCard[myField][i][1]);
										if (cardbookHTMLTypes.isMyCodePresent(myField, type, ABTypeFormat, myInputTypes)) {
											if (aIncludePref && cardbookHTMLUtils.getPrefBooleanFromTypes(aCard[myField][i][1])) {
												result.push(cardbookHTMLUtils.prefCSVPrefix + aCard[myField][i][0][subField]);
											} else {
												result.push(aCard[myField][i][0][subField]);
											}
										}
									}
								}
							}
						}
					} else {
						let type = myFieldArray.slice(1).join("_");
						if (type == "all") {
							if (aCard[myField]) {
								for (var i = 0; i < aCard[myField].length; i++) {
									let address = cardbookHTMLUtils.formatAddress(aCard[myField][i][0]);
									if (address) {
										result.push(address);
									}
								}
							}
						} else if (type == "notype") {
							if (aCard[myField]) {
								for (var i = 0; i < aCard[myField].length; i++) {
									let inputTypes = cardbookHTMLUtils.getOnlyTypesFromTypes(aCard[myField][i][1]);
									let emptyType = inputTypes.length == 0 || inputTypes[0] == "";
									if (emptyType && aCard[myField][i][3].length == 0 && aCard[myField][i][2] == "") {
										let address = cardbookHTMLUtils.formatAddress(aCard[myField][i][0]);
										if (address) {
											result.push(address);
										}
									}
								}
							}
						} else {
							for (var i = 0; i < aCard[myField].length; i++) {
								let found = false;
								let address = cardbookHTMLUtils.formatAddress(aCard[myField][i][0]);
								if (address) {
									if (aCard[myField][i][3].length != 0 && aCard[myField][i][2] != "") {
										for (var j = 0; j < aCard[myField][i][3].length; j++) {
											let tmpArray = aCard[myField][i][3][j].split(":");
											if (tmpArray[0] == "X-ABLABEL" && tmpArray[1].replaceAll(" ", "_") == type) {
												found = true;
												break;
											}
										}
									}
									if (found) {
										if (aIncludePref && cardbookHTMLUtils.getPrefBooleanFromTypes(aCard[myField][i][1])) {
											result.push(cardbookHTMLUtils.prefCSVPrefix + address);
										} else {
											result.push(address);
										}
									} else {
										var ABType = cardbookNewPreferences.getType(aCard.dirPrefId);
										var ABTypeFormat = cardbookHTMLUtils.getABTypeFormat(ABType);
										var myInputTypes = cardbookHTMLUtils.getOnlyTypesFromTypes(aCard[myField][i][1]);
										if (cardbookHTMLTypes.isMyCodePresent(myField, type, ABTypeFormat, myInputTypes)) {
											if (aIncludePref && cardbookHTMLUtils.getPrefBooleanFromTypes(aCard[myField][i][1])) {
												result.push(cardbookHTMLUtils.prefCSVPrefix + address);
											} else {
												result.push(address);
											}
										}
									}
								}
							}
						}
					}
				}
			} else {
				var myType = myFieldArray.slice(2).join("_");
				if (myType == "all") {
					if (aCard[myField]) {
						for (var i = 0; i < aCard[myField].length; i++) {
							if (aCard[myField][i][0][myPosition] != "") {
								result.push(aCard[myField][i][0][myPosition]);
							}
						}
					}
				} else if (myType == "array") {
					if (aCard[myField].length != 0) {
						result = result.concat(aCard[myField]);
					}
				} else {
					if (aCard[myField]) {
						var ABType = cardbookNewPreferences.getType(aCard.dirPrefId);
						var ABTypeFormat = cardbookHTMLUtils.getABTypeFormat(ABType);
						for (var i = 0; i < aCard[myField].length; i++) {
							if (myType == "notype") {
								let inputTypes = cardbookHTMLUtils.getOnlyTypesFromTypes(aCard[myField][i][1]);
								let emptyType = inputTypes.length == 0 || inputTypes[0] == "";
								if (emptyType && aCard[myField][i][3].length == 0 && aCard[myField][i][2] == "") {
									result.push(aCard[myField][i][0][myPosition]);
								}
							} else {
								let found = false;
								if (aCard[myField][i][3].length != 0 && aCard[myField][i][2] != "") {
									for (var j = 0; j < aCard[myField][i][3].length; j++) {
										let tmpArray = aCard[myField][i][3][j].split(":");
										if (tmpArray[0] == "X-ABLABEL" && tmpArray[1].replaceAll(" ", "_") == myType) {
											found = true;
											break;
										}
									}
								}
								if (found) {
									if (aIncludePref && cardbookHTMLUtils.getPrefBooleanFromTypes(aCard[myField][i][1])) {
										result.push(cardbookHTMLUtils.prefCSVPrefix + aCard[myField][i][0][myPosition]);
									} else {
										result.push(aCard[myField][i][0][myPosition]);
									}
								} else {
									var myInputTypes = cardbookHTMLUtils.getOnlyTypesFromTypes(aCard[myField][i][1]);
									if (cardbookHTMLTypes.isMyCodePresent(myField, myType, ABTypeFormat, myInputTypes)) {
										if (aIncludePref && cardbookHTMLUtils.getPrefBooleanFromTypes(aCard[myField][i][1])) {
											result.push(cardbookHTMLUtils.prefCSVPrefix + aCard[myField][i][0][myPosition]);
										} else {
											result.push(aCard[myField][i][0][myPosition]);
										}
									}
								}
							}
						}
					}
				}
			}
		} else if (aField == "age") {
			result.push(cardbookHTMLDates.getAge(aCard));
		} else if (aField == "ABName") {
			result.push(cardbookNewPreferences.getName(aCard.dirPrefId));
		} else {
			if (aCard[aField]) {
				result.push(aCard[aField]);
			} else {
				for (var i = 0; i < aCard.others.length; i++) {
					let delim = aCard.others[i].indexOf(":", 0);
					let header = aCard.others[i].substr(0, delim);
					let value = aCard.others[i].substr(delim+1, aCard.others[i].length);
					let headerTmp = header.split(";");
					if (aField == headerTmp[0]) {
						result.push(value);
						break;
					}
				}
			}
		}
		return result;
	},

};
